from typing import Any, Generic, List, Optional, Type, TypeVar, Union

from django.core.exceptions import ValidationError
from django.db.models import JSONField
from nwon_baseline.pydantic import pydantic_model_to_dict
from nwon_baseline.typings import AnyDict
from pydantic import BaseModel
from pydantic import ValidationError as PydanticValidationError

PydanticModel = TypeVar("PydanticModel", bound=BaseModel)


class PydanticJsonField(Generic[PydanticModel], JSONField):
    """
    An extension of a JSONField that uses pydantic models for data validation.

    If data is valid for multiple given models the first valid model is returned.
    """

    pydantic_model: Type[PydanticModel]

    def __init__(self, *args, pydantic_model: Type[PydanticModel], **kwargs):
        self.pydantic_model = pydantic_model
        kwargs.pop("pydantic_model", None)
        super().__init__(*args, **kwargs)

    def __get__(self, instance, owner) -> PydanticModel:  # type: ignore
        """Ensure correct type hinting when accessing the field on a model instance."""
        return super().__get__(instance, owner)  # type: ignore

    def validate(self, value: Optional[BaseModel], model_instance):
        if value is None:
            super().validate(value, model_instance)
        else:
            super().validate(pydantic_model_to_dict(value), model_instance)

        self._validate_schema(value)

    def from_db_value(self, value, expression, connection) -> Optional[PydanticModel]:
        """
        Converts the database value into a Pydantic model instance when retrieving it.
        """

        if value is None:
            return value

        # Validate and return the parsed Pydantic model
        return self._validate_schema(value)

    def to_python(self, value: Any) -> Optional[PydanticModel]:
        """
        Converts the JSON value to a Pydantic model instance in Python.
        """
        if value is None:
            return value

        return self._validate_schema(value)

    def get_prep_value(self, value):
        """Ensure that the value stored in the DB is a JSON serializable dictionary."""

        if isinstance(value, self.pydantic_model):
            return pydantic_model_to_dict(value)
        if isinstance(value, dict):
            return value  # Already a dictionary

        raise ValueError(f"Cannot store {type(value)} in {self.name}")

    def _validate_schema(
        self, value: Union[str, AnyDict, BaseModel, None]
    ) -> Optional[PydanticModel]:
        """
        Validates the value against the provided Pydantic models.
        Returns the first valid Pydantic model instance or raises ValidationError.
        """

        # Skip validation during fake migrations
        if self.model.__module__ == "__fake__":
            return value  # type: ignore

        if value is None:
            return None

        errors: List[str] = []

        try:
            if isinstance(value, str):
                return self.pydantic_model.model_validate_json(value)
            else:
                return self.pydantic_model.model_validate(value)

        except PydanticValidationError as exc:
            errors.append(
                f"JSON does not fit Pydantic model {self.pydantic_model.__name__} {format(exc)}"
            )

        # Raise a validation error if no model matches
        raise ValidationError(
            f"Value does not match any Pydantic model: {errors}", code="invalid"
        )
